<!doctype html>
<html lang="en">
  <head>
    <meta charset="utf-8">
    <meta name="viewport" content="width=device-width, initial-scale=1, shrink-to-fit=no">

    <link rel="stylesheet" href="webjars/bootstrap/4.3.1/css/bootstrap.min.css">
    <link rel="stylesheet" href="webjars/open-iconic/1.1.1/font/css/open-iconic-bootstrap.css">

    <link rel="icon" type="image/png" sizes="32x32" href="favicon-32x32.png">
    <link rel="icon" type="image/png" sizes="16x16" href="favicon-16x16.png">
    <link rel="shortcut icon" href="favicon.ico">
    <meta name="theme-color" content="#ffffff">

    <style>
      .auto-index tbody tr:first-child {
        counter-reset: AutoIndexCounter;
      }

      .auto-index tbody tr th:first-child:before {
        counter-increment: AutoIndexCounter;
        content: counter(AutoIndexCounter);
      }
    </style>

    <title>Key/Value - Skein | {{appId}}</title>
  </head>

  <body>
    <nav class="navbar navbar-expand-md navbar-dark bg-dark">
      <span class="navbar-brand">
        <img src="yarn_white.svg" width="28" height="28" class="d-inline-block align-top" alt="">
        Skein
      </span>

      <button class="navbar-toggler" type="button" data-toggle="collapse" data-target="#navbarToggler" aria-controls="navbarToggler" aria-expanded="false" aria-label="Toggle navigation">
        <span class="navbar-toggler-icon"></span>
      </button>

      <div class="collapse navbar-collapse" id="navbarToggler">
        <ul class="navbar-nav mr-auto">
          <li class="nav-item">
            <a class="nav-link" href="./overview">Overview</a>
          </li>
          <li class="nav-item active">
            <a class="nav-link" href="./kv">Key/Value</a>
          </li>
          <li class="nav-item dropdown">
            <a class="nav-link dropdown-toggle {{^pages}}disabled{{/pages}}" {{^pages}}title="No service pages available"{{/pages}} href="#" id="servicePages" data-toggle="dropdown" aria-haspopup="true" aria-expanded="false">
              Service Pages
            </a>
            <div class="dropdown-menu" aria-labelledby="servicePages">
              {{#pages}}
              <a class="dropdown-item" href=".{{proxyPrefix}}/{{getValue}}">{{getKey}}</a>
              {{/pages}}
            </div>
          </li>
        </ul>
        <div class="dropdown-divider"></div>
        <span class="navbar-text" id="application">{{appId}}</span>
      </div>
    </nav>

    <div class="table-responsive">
      <table class="table table-striped table-bordered table-sm auto-index">
        <colgroup>
          <col style="width: 5%;">
          <col style="width: 20%;">
          <col style="width: 75%;">
        </colgroup>
        <thead>
          <tr>
            <th scope="col">#</th>
            <th scope="col">Key</th>
            <th scope="col">Value</th>
          </tr>
        </thead>
        <tbody>
          {{#kv}}
          <tr>
            <th scope="row"></th>
            <td>{{getKey}}</td>
            <td>{{getValue}}</td>
          </tr>
          {{/kv}}
        </tbody>
      </table>
    </div>

    <script src="webjars/jquery/3.4.1/jquery.min.js"></script>
    <script src="webjars/bootstrap/4.3.1/js/bootstrap.bundle.min.js"></script>
  </body>
</html>
